#pragma once
#ifndef TessellationKhouryEtAl2019_h
#define TessellationKhouryEtAl2019_h

#include "../AbstractTessellation.h"
#include "../Program.h"

//=================================================================================================
/**
*	This class implements Adaptive GPU Tessellation with Compute Shaders as described by Jad
*	Khoury, Jonathan Dupuy, and Christophe Riccio. Adaptive GPU tessellation with compute shaders.
*	In GPU Zen 2: Advanced Rendering Techniques, W. Engel, Ed. Black Cat Publishing, 2019, ch. 1,
*	pp. 316.
*
*	@author	Daniel Cornel
**/
//=================================================================================================
class TessellationKhouryEtAl2019 : public AbstractTessellation
{
public:
	/// Constructor
	TessellationKhouryEtAl2019(const Heightfield& heightfield);
	/// Destructor
	~TessellationKhouryEtAl2019() override;
	/// Performs a per-frame update of the tessellation
	void update(const Camera& camera) override;
	/// Renders the geometry
	void render(const Camera& camera) override;

private:
	/// Performs one subdivision/merge iteration
	void subdivideMerge(const Camera& camera);
	/// Resets the temporary buffers and creates the command for the next update dispatch
	void resetBuffers();

public:
	GLuint mTriangleInstanceVertexBuffer = 0u;
	GLuint mTriangleInstanceIndexBuffer = 0u;
	GLuint mGridVertexBuffer = 0u;
	GLuint mGridIndexBuffer = 0u;
	GLuint mSubdivisionBuffer1 = 0u;
	GLuint mSubdivisionBuffer2 = 0u;
	GLuint mCulledTrianglesBuffer = 0u;
	GLuint mNumInputTrianglesBuffer = 0u;
	GLuint mNumOutputTrianglesBuffer = 0u;
	GLuint mIndirectDispatchBuffer = 0u;
	GLuint mIndirectDrawBuffer = 0u;
	GLuint mChangesAtomicCounterBuffer = 0u;
	GLuint mVAO = 0u;
	bool mPingPong = false;

	std::unique_ptr<Program> mTerrainProgram = nullptr;
	std::unique_ptr<Program> mLoDProgram = nullptr;
	std::unique_ptr<Program> mIndirectProgram = nullptr;
};

#endif